function [Vhat, pval, pval_refine, output] = jpe_ConstantBetaTest(Ycont,Zcont,Yjump,Zjump,deltan,rbar,Nsimul,k,lossfun)

%% Notes
% Input
%
% Ycont:  N x kn x 4, diffusive returns of Y for spot vol estimation 
%         indexed by q in 4 windows {tau-,tau+,eta-,eta+}.
% Zcont:  kn x 4, diffusive returns of Z index by q in {tau-,tau+,eta-,eta+}
% Yjump: N x 2, jumps in Y
% Zjump: 1 x 2, jumps in Z
% deltan: sampling interval
% rbar: the upper bound of number of factors
% Nsimul: number of simulations for generating CV
% k: mixed scale
% lossfun: loss function to be used (set [] if using the default loss function)


% Output
% Vhat: the test stat
% pval: p value
% pval_refine: refined p value
% output: collecting all relevant output, can be printed via jpe_print(output)


%% Code begins here

% Step 1. 
% Estimate spot beta [beta: N x 2] and test stat
if isempty(lossfun)
    lossfun = @(x)(min(abs(x),x.^2)); % <- Change this if using other default loss function
end
[N,kn,junk] = size(Ycont);
beta = Yjump ./ repmat(Zjump,N,1); % spot betas [N x 2]
Vhat = mean(lossfun(beta(:,1) - beta(:,2))); % test stat

% Step 2. Generate residuals Xhat(q), q in windows {tau-,tau+,eta-,eta+}
beta_avg = mean(beta,2); % use the average beta as pre-estimator, N x 1
% Generate Xhat
Xhat = zeros(N,kn,4);
for q = 1:4
    Xhat(:,:,q) = (Ycont(:,:,q) - beta_avg * Zcont(:,q)') / sqrt(deltan); 
end

% Step 3. Estimate the M matrices
% Compute Lambda(q)
Lhat = zeros(N,rbar,4); % \hat{Lambda}
Fhat = zeros(kn,rbar,4); %\hat{F}
LZhat = zeros(rbar,4); %\hat{lambda}_Z
for q = 1:4
    [Lhat(:,:,q),Fhat(:,:,q)] = jpe_Lhat(Xhat(:,:,q),rbar); 
    % <- calls jpe_Lhat function for getting \hat{Lambda} and \hat{F}
    LZhat(:,q) = Fhat(:,:,q)' * Zcont(:,q) / kn; % Estimate of Lambda_Z
end
% Compute \hat{M}_{Cn}(p,q), rbar x rbar x 4 x 4
MChat = zeros(rbar,rbar,4,4);
for p = 1:4
    for q = 1:4
        MChat(:,:,p,q) = Lhat(:,:,p)' * Lhat(:,:,q) / N;
    end
end

% Compute \hat{M}_{\epsilon}(q), 4 x 1
Mepshat = zeros(4,1);
for q = 1:4
    tmpXhat = Xhat(:,:,q);
    Mepshat(q) = trace(tmpXhat' * tmpXhat) / kn / N - trace(MChat(:,:,q,q));
end

% Step 4. Simulation
Vsimul = zeros(Nsimul,1); % Simulated test stat in Algo 1
Vsimul_refine = zeros(Nsimul,1); % Simulated test stat in Algo 2

parfor mc = 1:Nsimul
    kappa_tau = rand(1);
    kappa_eta = rand(1);
    w = [sqrt(kappa_tau) / Zjump(1), sqrt(k-kappa_tau) / Zjump(1),...
         sqrt(kappa_eta) / Zjump(2), sqrt(k-kappa_eta) / Zjump(2)];
    zeta_tilde = randn(rbar,4);
    
    % w and zeta_tilde stores the corresponding value for {tau-,tau+,eta-,eta+}
    
    A_tau = 0;
    A_eta = 0;
    B_taueta = 0;
    
    % getting A_tau
    for p = 1:2 % {tau-,tau+}
        for q = 1:2
            A_tau = A_tau + w(p) * w(q) * zeta_tilde(:,p)' ...
                * MChat(:,:,p,q) * zeta_tilde(:,q);
        end
        A_tau = A_tau + w(p)^2 * Mepshat(p);
    end
    
    % getting A_eta
    for p = 3:4 % {eta-,eta+}
        for q = 3:4
            A_eta = A_eta + w(p) * w(q) * zeta_tilde(:,p)' ...
                * MChat(:,:,p,q) * zeta_tilde(:,q);
        end
        A_eta = A_eta + w(p)^2 * Mepshat(p);
    end

    % getting B_{tau,eta}
    for p = 1:2 % {tau-,tau+}
        for q = 3:4 % {eta-,eta+}
            B_taueta = B_taueta + w(p) * w(q) * zeta_tilde(:,p)'...
                * MChat(:,:,p,q) * zeta_tilde(:,q);
        end
    end
    
    L = A_tau + A_eta - 2 * B_taueta;
    Vsimul(mc) = L;
    
    % Compute refinement
    Psi_tau = 0; Psi_eta = 0;
    for q = 1:2
        Psi_tau = Psi_tau - w(q) * LZhat(:,q)' * zeta_tilde(:,q);
    end
    for q = 3:4
        Psi_eta = Psi_eta - w(q) * LZhat(:,q)' * zeta_tilde(:,q);
    end
    H = 2 * (A_tau * Psi_tau + A_eta * Psi_eta) - 2 * B_taueta * (Psi_tau + Psi_eta);
    Vsimul_refine(mc) = L + H * sqrt(deltan);
end
pval = mean(Vsimul > Vhat/deltan);
cv90 = prctile(Vsimul,90);
cv95 = prctile(Vsimul,95);
cv99 = prctile(Vsimul,99);

pval_refine = mean(Vsimul_refine > Vhat/deltan);
cv90_refine = prctile(Vsimul_refine,90);
cv95_refine = prctile(Vsimul_refine,95);
cv99_refine = prctile(Vsimul_refine,99);

output.Vhat = Vhat;
output.pval = pval;
output.pval_refine = pval_refine;

output.Nsimul = Nsimul;
output.rbar = rbar;
output.deltan = deltan;
output.k = k;

output.beta = beta;
output.MChat = MChat;
output.Mepshat = Mepshat;

output.Vsimul = Vsimul;
output.cv90 = cv90;
output.cv95 = cv95;
output.cv99 = cv99;

output.Vsimul_refine = Vsimul_refine;
output.cv90_refine = cv90_refine;
output.cv95_refine = cv95_refine;
output.cv99_refine = cv99_refine;

output.type = 'ConstantBetaTest';
output.lossfun = lossfun;

