function output = jpe_joint_test(jpedata,deltan,rbar,Nsimul,lossfun)

%% Notes
% Input
%
% jpedata structure:
%
% Ycont:  N x kn x (2T), diffusive returns of Y for spot vol estimation 
%         indexed by q in (2T) windows {tau1-, tau1+, tau2-, tau2+, ...}.
% Zcont:  kn x (2T), diffusive returns of Z index by q in {tau1-,tau1+, tau2-, tau2+, ...}.
% Yjump: N x T, jumps in Y
% Zjump: 1 x T, jumps in Z
% kmix: mixed scale
%
% deltan: sampling interval
% rbar: the upper bound of number of factors
% Nsimul: number of simulations for generating critical values
% lossfun: loss function to be used (set [] if using the default loss function)


% Output

%% Data Unload
Ycont = jpedata.Ycont;
Zcont = jpedata.Zcont;
Yjump = jpedata.Yjump;
Zjump = jpedata.Zjump;
k = jpedata.kmix;


%% Code begins here
output.message_setting = 'Settings:';
if isempty(lossfun)
    lossfun = @(x)(min(abs(x),x.^2)); % <- Change this if using other default loss function
end
output.lossfun = lossfun;
output.mixscale = k;
output.Nsimul = Nsimul;
output.rbar = rbar;
output.deltan = deltan;
output.type = 'ConstantBetaTest_MultipleTimes';

%% Step 1. Estimate spot beta [beta: N x T]
[N,kn,T2] = size(Ycont);
T = T2/2; % Number of jumps
beta = Yjump ./ repmat(Zjump,N,1); % spot betas [N x T]

output.message_samplesize = 'Sample size information:';
output.N = N;
output.kn = kn;
output.number_of_jumps = T;


%% Step 2. Compute test stats

Vhat = zeros(T-1,1);
vhat_raw = zeros(T-1,1); % vhat without correction
bias = zeros(T-1,1);
diffus_var = zeros(T-1,1); % used to produce relative vhat measure
jump_var = zeros(T-1,1);   % used to produce relative vhat measure

Vsimul = zeros(T-1,Nsimul);

% Simulate all kappa and zeta random variables (so as to capture joint dependence)
kappa_all = rand(T,Nsimul);
zeta_all  = randn(rbar,2*T,Nsimul);

for t = 1:(T-1)
    % Compute for consecutive jump pairs
    beta_eta = beta(:,t);
    beta_tau = beta(:,t+1);
    beta_avg = (beta_eta + beta_tau) / 2; % use the average beta as pre-estimator, N x 1        
    % Compute test stat and raw estimate of vhat
        Vhat(t) = mean(lossfun(beta_tau - beta_eta)); % test stats
        vhat_raw(t) = mean((beta_tau - beta_eta).^2); % raw vhat
        
    % Prep
        Zjump_eta = Zjump(t);
        Zjump_tau = Zjump(t+1);
        Yjump_eta = Yjump(:,t);
        Yjump_tau = Yjump(:,t+1);

    % Generate Xhat
        Xhat = zeros(N,kn,4);
        bias_vec = zeros(4,1);
        diffus_var_vec = zeros(4,1);
        for q = 1:4
            tmpYcont = Ycont(:,:,(t-1)*2+q);
            tmpZcont = Zcont(:,(t-1)*2+q);
            tmpXhat = (tmpYcont - beta_avg * tmpZcont') / sqrt(deltan); % N x kn
            Xhat(:,:,q) = tmpXhat;
            bias_vec(q) = mean(tmpXhat(:).^2);
            diffus_var_vec(q) = mean(tmpYcont(:).^2);
        end
        % Bias term
        bias(t) = ((1/2) * bias_vec(1) + (k - 1/2) * bias_vec(2)) / Zjump_eta^2 + ...
                  ((1/2) * bias_vec(3) + (k - 1/2) * bias_vec(4)) / Zjump_tau^2;
        diffus_var(t) = (1/2) * diffus_var_vec(1) + (k - 1/2) * diffus_var_vec(2) + ...
                  (1/2) * diffus_var_vec(3) + (k - 1/2) * diffus_var_vec(4);
        % jump variations
        jump_var(t) = mean(Yjump_eta.^2) + mean(Yjump_tau.^2);
    % Estimate the M matrices
        % Compute Lambda(q)
        Lhat = zeros(N,rbar,4); % \hat{Lambda}
        Fhat = zeros(kn,rbar,4); %\hat{F}
        LZhat = zeros(rbar,4); %\hat{lambda}_Z
        for q = 1:4
            [Lhat(:,:,q),Fhat(:,:,q)] = jpe_Lhat(Xhat(:,:,q),rbar); 
            % <- calls jpe_Lhat function for getting \hat{Lambda} and \hat{F}
            LZhat(:,q) = Fhat(:,:,q)' * Zcont(:,q) / kn; % Estimate of Lambda_Z
        end
        % Compute \hat{M}_{Cn}(p,q), rbar x rbar x 4 x 4
        MChat = zeros(rbar,rbar,4,4);
        for p = 1:4
            for q = 1:4
                MChat(:,:,p,q) = Lhat(:,:,p)' * Lhat(:,:,q) / N;
            end
        end

        % Compute \hat{M}_{\epsilon}(q), 4 x 1
        Mepshat = zeros(4,1);
        for q = 1:4
            tmpXhat = Xhat(:,:,q);
            Mepshat(q) = trace(tmpXhat' * tmpXhat) / kn / N - trace(MChat(:,:,q,q));
        end

    %Simulation
        parfor mc = 1:Nsimul
            kappa_eta = kappa_all(t,mc);
            kappa_tau = kappa_all(t+1,mc);
            w = [sqrt(kappa_eta) / Zjump(t),   sqrt(k-kappa_eta) / Zjump(t),...
                 sqrt(kappa_tau) / Zjump(t+1), sqrt(k-kappa_tau) / Zjump(t+1)];
            zeta_tilde = zeta_all(:,((t-1)*2+1):((t-1)*2+4),mc); 
            % w and zeta_tilde stores the corresponding value for {tau-,tau+,eta-,eta+}
            A_eta = 0;
            A_tau = 0;
            B_taueta = 0;
            % getting A_eta
            for p = 1:2 % {eta-,eta+}
                for q = 1:2
                    A_eta = A_eta + w(p) * w(q) * zeta_tilde(:,p)' ...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
                A_eta = A_eta + w(p)^2 * Mepshat(p);
            end

            % getting A_tau
            for p = 3:4 % {tau-,tau+}
                for q = 3:4
                    A_tau = A_tau + w(p) * w(q) * zeta_tilde(:,p)' ...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
                A_tau = A_tau + w(p)^2 * Mepshat(p);
            end

            % getting B_{tau,eta}
            for p = 1:2 % {tau-,tau+}
                for q = 3:4 % {eta-,eta+}
                    B_taueta = B_taueta + w(p) * w(q) * zeta_tilde(:,p)'...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
            end

            L = A_tau + A_eta - 2 * B_taueta;
            Vsimul(t,mc) = L;

        end
end

%% Step 3. Compute p values
% Vhat: pairwise test stats [T-1 x 1]
% bias: bias terms (estimates for centers of Vhat) [T-1 x 1]
% Vsimul: simulated limiting distribution of Vhat [T-1 x MC]
    
%% Joint tests    

Vhat_scaled = Vhat ./ bias;
Vsimul_scaled   = Vsimul ./ repmat(bias,1,Nsimul);

Vhat_mean_scaled  = mean(Vhat_scaled);
pval_mean_scaled = mean(mean(Vsimul_scaled) > Vhat_mean_scaled / deltan);


%% output

% output.message_beta = 'Jump beta estimates:';
% output.beta = beta;
% output.Zjump_eta = Zjump(1:end-1)';
% output.Zjump_tau = Zjump(2:end)';

%output.message_pairwise = 'Pairwise test results:';
%output.Vhat = Vhat;
%output.Vsimul = Vsimul;

output.message_jointtest = 'P values for joint tests:';
output.pval_mean_scaled = pval_mean_scaled;
