function output = jpe_main(jpedata,deltan,rbar,Nsimul,wins,lossfun)

%% Notes
% Input
%
% jpedata structure:
%
% Ycont:  N x kn x (2T), diffusive returns of Y for spot vol estimation 
%         indexed by q in (2T) windows {tau1-, tau1+, tau2-, tau2+, ...}.
% Zcont:  kn x (2T), diffusive returns of Z index by q in {tau1-,tau1+, tau2-, tau2+, ...}.
% Yjump: N x T, jumps in Y
% Zjump: 1 x T, jumps in Z
% kmix: mixed scale
%
% deltan: sampling interval
% rbar: the upper bound of number of factors
% Nsimul: number of simulations for generating critical values
% lossfun: loss function to be used (set [] if using the default loss function)


% Output

%% Data Unload
Ycont = jpedata.Ycont;
Zcont = jpedata.Zcont;
Yjump = jpedata.Yjump;
Zjump = jpedata.Zjump;
dates = jpedata.dates;
times = jpedata.times;
k = jpedata.kmix;
trunc = jpedata.trunc;


%% Code begins here
output.message_setting = 'Settings:';
if isempty(lossfun)
    lossfun = @(x)(min(abs(x),x.^2)); % <- Change this if using other default loss function
end
output.lossfun = lossfun;
output.mixscale = k;
output.Nsimul = Nsimul;
output.rbar = rbar;
output.winsorize = wins;
output.deltan = deltan;
output.type = 'ConstantBetaTest_MultipleTimes';

%% Step 1. Estimate spot beta [beta: N x T]
[N,kn,T2] = size(Ycont);
T = T2/2; % Number of jumps
beta = Yjump ./ repmat(Zjump,N,1); % spot betas [N x T]

output.message_samplesize = 'Sample size information:';
output.N = N;
output.kn = kn;
output.number_of_jumps = T;


%% Step 2. Compute test stats

Vhat = zeros(T-1,1);
vhat_raw = zeros(T-1,1); % vhat without correction
bias = zeros(T-1,1);
diffus_var = zeros(T-1,1); % used to produce relative vhat measure
jump_var = zeros(T-1,1);   % used to produce relative vhat measure

Vsimul = zeros(T-1,Nsimul);
Vsimul_refine = zeros(T-1,Nsimul);

% Simulate all kappa and zeta random variables (so as to capture joint dependence)
kappa_all = rand(T,Nsimul);
zeta_all  = randn(rbar,2*T,Nsimul);

for t = 1:(T-1)
    % Compute for consecutive jump pairs
    beta_eta = beta(:,t);
    beta_tau = beta(:,t+1);
    beta_avg = (beta_eta + beta_tau) / 2; % use the average beta as pre-estimator, N x 1        
    % Compute test stat and raw estimate of vhat
        beta_quant = quantile(abs(beta_tau - beta_eta),1-wins);  
        input_betas = min(abs(beta_tau - beta_eta),beta_quant);
        Vhat(t) = mean(lossfun(input_betas)); % test stats
        vhat_raw(t) = mean(input_betas.^2); % raw vhat
        
    % Prep
        Zjump_eta = Zjump(t);
        Zjump_tau = Zjump(t+1);
        Yjump_eta = Yjump(:,t);
        Yjump_tau = Yjump(:,t+1);

    % Generate Xhat
        Xhat = zeros(N,kn,4);
        bias_vec = zeros(4,1);
        diffus_var_vec = zeros(4,1);
        for q = 1:4
            tmpYcont = Ycont(:,:,(t-1)*2+q);
            tmpZcont = Zcont(:,(t-1)*2+q);
            if q<=2
                tmpYcont = min(tmpYcont,2*trunc(t)); tmpYcont = max(tmpYcont,-2*trunc(t));
                tmpZcont = min(tmpZcont,trunc(t)); tmpZcont = max(tmpZcont,-trunc(t));
            else
                tmpYcont = min(tmpYcont,2*trunc(t+1)); tmpYcont = max(tmpYcont,-2*trunc(t+1));
                tmpZcont = min(tmpZcont,trunc(t+1)); tmpZcont = max(tmpZcont,-trunc(t+1));
            end
            tmpXhat = (tmpYcont - beta_avg * tmpZcont') / sqrt(deltan); % N x kn
            Xhat(:,:,q) = tmpXhat;
            bias_vec(q) = mean(tmpXhat(:).^2);
            diffus_var_vec(q) = mean(tmpYcont(:).^2);
        end
        % Bias term
        bias(t) = ((1/2) * bias_vec(1) + (k - 1/2) * bias_vec(2)) / Zjump_eta^2 + ...
                  ((1/2) * bias_vec(3) + (k - 1/2) * bias_vec(4)) / Zjump_tau^2;
        diffus_var(t) = (1/2) * diffus_var_vec(1) + (k - 1/2) * diffus_var_vec(2) + ...
                  (1/2) * diffus_var_vec(3) + (k - 1/2) * diffus_var_vec(4);
        % jump variations
        jump_var(t) = mean(Yjump_eta.^2) + mean(Yjump_tau.^2);
    % Estimate the M matrices
        % Compute Lambda(q)
        Lhat = zeros(N,rbar,4); % \hat{Lambda}
        Fhat = zeros(kn,rbar,4); %\hat{F}
        LZhat = zeros(rbar,4); %\hat{lambda}_Z
        for q = 1:4
            [Lhat(:,:,q),Fhat(:,:,q)] = jpe_Lhat(Xhat(:,:,q),rbar); 
            % <- calls jpe_Lhat function for getting \hat{Lambda} and \hat{F}
            LZhat(:,q) = Fhat(:,:,q)' * Zcont(:,q) / kn; % Estimate of Lambda_Z
        end
        % Compute \hat{M}_{Cn}(p,q), rbar x rbar x 4 x 4
        MChat = zeros(rbar,rbar,4,4);
        for p = 1:4
            for q = 1:4
                MChat(:,:,p,q) = Lhat(:,:,p)' * Lhat(:,:,q) / N;
            end
        end

        % Compute \hat{M}_{\epsilon}(q), 4 x 1
        Mepshat = zeros(4,1);
        for q = 1:4
            tmpXhat = Xhat(:,:,q);
            Mepshat(q) = trace(tmpXhat' * tmpXhat) / kn / N - trace(MChat(:,:,q,q));
        end

    %Simulation
        parfor mc = 1:Nsimul
            kappa_eta = kappa_all(t,mc);
            kappa_tau = kappa_all(t+1,mc);
            w = [sqrt(kappa_eta) / Zjump(t),   sqrt(k-kappa_eta) / Zjump(t),...
                 sqrt(kappa_tau) / Zjump(t+1), sqrt(k-kappa_tau) / Zjump(t+1)];
            zeta_tilde = zeta_all(:,((t-1)*2+1):((t-1)*2+4),mc); 
            % w and zeta_tilde stores the corresponding value for {tau-,tau+,eta-,eta+}
            A_eta = 0;
            A_tau = 0;
            B_taueta = 0;
            % getting A_eta
            for p = 1:2 % {eta-,eta+}
                for q = 1:2
                    A_eta = A_eta + w(p) * w(q) * zeta_tilde(:,p)' ...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
                A_eta = A_eta + w(p)^2 * Mepshat(p);
            end

            % getting A_tau
            for p = 3:4 % {tau-,tau+}
                for q = 3:4
                    A_tau = A_tau + w(p) * w(q) * zeta_tilde(:,p)' ...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
                A_tau = A_tau + w(p)^2 * Mepshat(p);
            end

            % getting B_{tau,eta}
            for p = 1:2 % {tau-,tau+}
                for q = 3:4 % {eta-,eta+}
                    B_taueta = B_taueta + w(p) * w(q) * zeta_tilde(:,p)'...
                        * MChat(:,:,p,q) * zeta_tilde(:,q);
                end
            end

            L = A_tau + A_eta - 2 * B_taueta;
            Vsimul(t,mc) = L;

            % Compute refinement
            Psi_eta = 0; Psi_tau = 0;
            for q = 1:2
                Psi_eta = Psi_eta - w(q) * LZhat(:,q)' * zeta_tilde(:,q);
            end
            for q = 3:4
                Psi_tau = Psi_tau - w(q) * LZhat(:,q)' * zeta_tilde(:,q);
            end
            H = 2 * (A_tau * Psi_tau + A_eta * Psi_eta) - 2 * B_taueta * (Psi_tau + Psi_eta);
            Vsimul_refine(t,mc) = L + H * sqrt(deltan);
        end
end

%% Step 3. Compute p values
% Vhat: pairwise test stats [T-1 x 1]
% bias: bias terms (estimates for centers of Vhat) [T-1 x 1]
% Vsimul: simulated limiting distribution of Vhat [T-1 x MC]
% Vsimul_refine: simulated refined limiting distribution of Vhat [T-1 x MC]

%% Pairwise tests
    pval = zeros(T-1,1);
    pval_refine = zeros(T-1,1);

    for t = 1:T-1
        pval(t) = mean(Vsimul(t,:) > Vhat(t)/deltan);
        pval_refine(t) = mean(Vsimul_refine(t,:) > Vhat(t)/deltan);
    end


    
%% Joint tests    
if T > 2
    dates = jpedata.dates; years = floor(dates/10^4); temp = diff(years); I = find(temp~=0);
    year_begin = [1 I+1]; year_end = [I length(years)];
    
    Vhat_scaled = Vhat ./ bias;
    Vsimul_scaled   = Vsimul ./ repmat(bias,1,Nsimul);
    
    nyears = length(year_begin); pval_mean_scaled = zeros(nyears,1);
    for i=1:nyears
        Vhat_mean_scaled  = mean(Vhat_scaled(year_begin(i):year_end(i)-1));
        pval_mean_scaled(i) = mean(mean(Vsimul_scaled(year_begin(i):year_end(i)-1,:)) > Vhat_mean_scaled / deltan);
    end
end

%% Confidence Interval for vhat
vhat = vhat_raw - deltan * bias;
Sigma_S = zeros(T-1,1);
vtilde_simul = zeros(T-1,Nsimul);
unitzeta_tau_minus = randn(T-1,Nsimul);
unitzeta_tau_plus  = randn(T-1,Nsimul);
theta = N * deltan;
vhatCI95 = zeros(T-1,2);
vhatCI90 = zeros(T-1,2);
for t = 1:(T-1)
    tmpvhat = vhat(t);
    beta_eta = beta(:,t);
    beta_tau = beta(:,t+1);

    Zcont_tauminus = Zcont(:,2*t+1);
    Zcont_tauplus  = Zcont(:,2*t+2);
    
    Zjump_tau = Zjump(t+1);
    
    % spot variance of Z at time tau
    spotvol_tauminus = mean(Zcont_tauminus.^2) / deltan;
    spotvol_tauplus  = mean(Zcont_tauplus.^2)  / deltan;
    
    % Sigma_S
    Sigma_S(t) = mean((input_betas).^4) - tmpvhat^2;
    
    % Simulate \tilde{v}_n^{(k)}
    Stilde = randn(1,Nsimul) * sqrt(Sigma_S(t));
    kappa_tau = kappa_all(t+1,:);
    vtilde_simul(t,:) = Stilde - 2 * sqrt(theta) * tmpvhat * (...
                 sqrt(kappa_tau)     ./ Zjump_tau .* sqrt(spotvol_tauminus) .* unitzeta_tau_minus(t,:) + ...
                 sqrt(k - kappa_tau) ./ Zjump_tau .* sqrt(spotvol_tauplus)  .* unitzeta_tau_plus(t,:));
             
    % Compute confidence intervals
    tmp_q95 = prctile(abs(vtilde_simul(t,:)),97.5);
    tmp_q90 = prctile(abs(vtilde_simul(t,:)),95);
    vhatCI95(t,:) = [tmpvhat - tmp_q95/sqrt(N), tmpvhat + tmp_q95/sqrt(N)];
    vhatCI90(t,:) = [tmpvhat - tmp_q90/sqrt(N), tmpvhat + tmp_q90/sqrt(N)];
end

vhat_mean = mean(vhat);
% vtilde_mean_simul = mean(vtilde_simul,1);
% tmp_q = prctile(abs(vtilde_mean_simul),97.5);
% vhat_mean_CI95 = [vhat_mean - tmp_q/sqrt(N-wins), vhat_mean + tmp_q/sqrt(N-wins)];


%% output

output.message_beta = 'Jump beta estimates:';
output.beta = beta;
output.Zjump_eta = Zjump(1:end-1)';
output.Zjump_tau = Zjump(2:end)';

output.message_pairwise = 'Pairwise test results:';
output.eta = [dates(1:end-1)' times(1:end-1)'];
output.tau = [dates(2:end)' times(2:end)'];
output.Vhat = Vhat;
output.pval = pval;
output.pval_refine = pval_refine;
output.Vsimul = Vsimul;
output.Vsimul_refine = Vsimul_refine;
if T == 2
    output.MChat = MChat;
    output.Mepshat = Mepshat;
end

output.message_vhat = 'Estimation results of vhat:';
output.vhat = vhat;
output.vhat_raw = vhat_raw;
output.bias = bias;
output.diffusive_variance = diffus_var;
output.jump_variance = jump_var;
output.D = diffus_var + jump_var;
output.vhatCI95 = vhatCI95;
output.vhatCI90 = vhatCI90;
output.vhat_mean = vhat_mean;
%output.vhat_mean_CI95 = vhat_mean_CI95;
output.Sigma_S = Sigma_S;
output.vtilde_simul = vtilde_simul;

if T > 2
    output.message_jointtest = 'P values for joint tests:';
    output.pval_mean_scaled = pval_mean_scaled;    
end
